/*
 * JRainWords 0.3 - Useful (?) educational (??) tool (???).
 *
 * JRainWords is a small Java program useful in teaching children how
 * to use a computer keyboard.
 *
 * Copyright 2001, 2002 Davide Alberani <alberanid@libero.it> 
 *
 * email:          alberanid@libero.it or alberanid@bigfoot.com
 * Jrw pages:      http://digilander.libero.it/alberanid/soft/jrw/
 * my homepage:    http://digilander.libero.it/alberanid/
 * ICQ UIN:        83641305 (nick 'Mad77')
 * Jabber ID:      alberanid@jabber.linux.it
 * PGP KeyID:      0x465BFD47 (the key is available in my homepage and
 *                                 through every keyserver)
 *
 * This code is released under the GPL license.
 *
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;

import java.io.*;
import java.util.*;
import java.net.URL;
import javax.swing.filechooser.*;

import com.sun.speech.freetts.Voice;
import com.sun.speech.freetts.audio.JavaClipAudioPlayer;
import com.sun.speech.freetts.en.us.CMULexicon;
import com.sun.speech.freetts.en.us.CMUDiphoneVoice;


// The main class.
public class Jrw extends JPanel {

	// Program name, version number and other infos.
	private final static String JRW_NAME = "JRainWords";
	private final static String JRW_VERSION = "0.3";
	private final static String JRW_LICENSE = "GPL";
	private final static String JRW_YEAR = "2001, 2002";
	private final static String JRW_AUTHOR = "Davide Alberani";
	private final static String JRW_EMAIL = "alberanid@libero.it";
	private final static String JRW_HOMEPAGE =
			"http://digilander.libero.it/alberanid/soft/jrw/";

	// The board.
	private static EditBoard editor;
	// Buttons bar.
	private JToolBar toolbar;
	// Menus.
	private JMenuBar menubar;
	// The label in the statusbar.
	private JLabel textlabel = new JLabel();

	// To choose a word, a color, etc.
	// XXX: initialize the seed?
	private Random rndgen = new Random();

	// Configuration file.
	private final static String jrwDirName = ".jrw";
	private final static String jrwConfFileName = "jrw.cfg";
	private static File confFile;

	// Some dialogs.
	private JFileChooser fileChooser;
	private JDialog wordsDialog;
	private JDialog optionsDialog;
	private JOptionPane messagesDialog = new JOptionPane();

	// XXX: use a Vector and put everything in a file + ColorChooser.
	private Color[] colorsList = {
		Color.blue,
		Color.green,
		Color.red,
		Color.magenta,
		Color.orange,
		new Color(100, 0, 150),
		new Color(35, 100, 80),
		new Color(160, 80, 50),
		new Color(160, 80, 255),
		new Color(255, 70, 255)
	};

	// Default properties, and options read from user's file.
	private static Properties defaultProperties = new Properties();
	private static Properties options = new Properties();
	static {
		defaultProperties.setProperty("wordsfile", "jrwwords.txt");
		defaultProperties.setProperty("fontsize", "16");
		defaultProperties.setProperty("speech", "true");
		defaultProperties.setProperty("windowheight", "480");
		defaultProperties.setProperty("windowwidth", "640");
		defaultProperties.setProperty("maxlength", "unlimited");
	}

	// The list of words from the external file.
	private Vector wordsList = new Vector();
	private Vector consideredWords = new Vector();

	// The word you have to guess.
	private String currentWord;

	// Max length choosing a new word (-1 means unlimited).
	private int maxLength = -1;

	// Default window size.
	private static int wh = 480;
	private static int ww = 640;

	// Commands for buttons and menu items.
	private Hashtable commands;

	// List of actions defined in this panel.
	private Action[] defaultActions = {
		new ExitAction(),
		new NewAction(),
		new ChangeWordAction(),
		new SaveAction(),   
		new OpenAction(),   
		new ModifyAction(),
		new ShowOptionsAction(),
		new AboutAction()
	};  

	// The speech synthesizer.
	private static Voice voice = new CMUDiphoneVoice();

	private static String[] congratulations = null;

	// Resources from the outher space.
	private static ResourceBundle resources;
	// Get the resources.
	static {
		try {
			resources = ResourceBundle.getBundle("resources.Jrw",
							 Locale.getDefault());
		} catch (MissingResourceException mre) {
			System.err.println
				("Unable to read resources/Jrw.properties");
			System.exit(1);
		}
	}


	// That's me!
	public Jrw() {
		super(true);
		initUserOptions();
		// Initialize the speech synthesizer.
		voice.setLexicon(new CMULexicon());
		JavaClipAudioPlayer ap = new JavaClipAudioPlayer();
		voice.getFeatures().setString(Voice.DATABASE_NAME,
				"cmu_kal/diphone_units16.bin");
		voice.setAudioPlayer(ap);
		voice.load();
		try {
			UIManager.setLookAndFeel(
				UIManager.getSystemLookAndFeelClassName());
		} catch (Exception exc) {
			System.err.println("Error loading Look&Feel: " + exc);
		}
		setBorder(BorderFactory.createEtchedBorder());
		setLayout(new BorderLayout());
		// The edit board.
		StyleContext sc = new StyleContext();
		DefaultStyledDocument doc = new DefaultStyledDocument(sc);
		editor = new EditBoard(doc, sc);
		commands = new Hashtable();
		Action[] actions = getActions();
		for (int i = 0; i < actions.length; i++) {
			Action a = actions[i];
			commands.put(a.getValue(Action.NAME), a);
		}
		JScrollPane scroller = new JScrollPane();
		scroller.setHorizontalScrollBarPolicy(
				JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		JViewport port = scroller.getViewport();
		port.add(editor);
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		JPanel statusbar = new JPanel();
		statusbar.add(textlabel);
		createMenubar();
		add("North", menubar);
		panel.add("North", createToolbar());
		panel.add("Center", scroller);
		add("Center", panel);
		add("South", statusbar);
		getCongratulations();
		initWords();
	}


	// Oh, well...
	public static void main(String[] args) {
		try {
			JFrame frame = new JFrame();
			frame.setTitle(JRW_NAME + " " + JRW_VERSION);
			frame.setBackground(Color.lightGray);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add("Center", new Jrw());
			frame.addWindowListener(new AppCloser());
			frame.pack();
			frame.setSize(ww, wh);
			// It's nice, isn't it?
			frame.show();
		} catch (Throwable t) {
			System.err.println("Serious trouble: " + t);
			t.printStackTrace();
		}
		getEditor().requestFocus();
	}


	// What to do on exit.
	private static final class AppCloser extends WindowAdapter {
		public void windowClosing(WindowEvent e) {
			// XXX: ask?
			System.exit(0);
		}
	}

	// The board you write into.
	private class EditBoard extends JTextPane implements KeyListener {

		DefaultStyledDocument doc;
		StyleContext styles;
		// There is new text in the board.
		boolean modified = false;

		public EditBoard(DefaultStyledDocument d, StyleContext sc) {
			addKeyListener(this);
			initDocument(d, sc);
		}
		public void initDocument(DefaultStyledDocument d,
					StyleContext sc) {
			this.doc = d;
			this.styles = sc;
			setStyledDocument(doc);
			setFontSize();
			modified = false;
		}
		public void setFontSize() {
			Style s = styles.addStyle(null, null);
			int fs = Integer.parseInt(
					options.getProperty("fontsize", "16"));
			if (fs <= 0) fs = 16;
			StyleConstants.setFontSize(s, fs);
			doc.setLogicalStyle(
					doc.getEndPosition().getOffset(), s);
		}
		public boolean wasModified() {
			return modified;
		}
		public void setModified(boolean m) {
			modified = m;
		}
		public void keyTyped(KeyEvent e) {
		}
		public void keyPressed(KeyEvent e) {
		}
		// Check performed every time a key is released.
		// XXX: optimize for speed.
		public void keyReleased(KeyEvent e) {
			char thechar = e.getKeyChar();
			if (thechar == ' ' || thechar == '\t' ||
					thechar == '\n' || thechar == '\r' ||
					thechar == '\f' ||
					thechar == KeyEvent.CHAR_UNDEFINED) {
				return;
			}
			setModified(true);
			String txt = "";
			String[] txttok = null;
			int caretpos = getCaretPosition();
			try {
				txt = getText(0, caretpos);
			} catch (BadLocationException ble) {
				messagesDialog.showMessageDialog(getFrame(),
						"No caret position",
						"Warning",
						JOptionPane.ERROR_MESSAGE);
				System.exit(20);
			}
			txttok = tokenize(txt);
			if (txttok.length == 0) {
				return;
			}
			String lastword = txttok[txttok.length - 1];
			// Got a match.
			if (lastword.equalsIgnoreCase(currentWord)) {
				gotcha(caretpos, lastword);
			}
		}
		// Wow, good shot!
		public void gotcha(int cp, String s) {
				// Change the color for the matched word.
				int len = s.length();
				Style st = styles.addStyle(null,
							getLogicalStyle());
				int r = rndgen.nextInt(colorsList.length);
				StyleConstants.setForeground(st, colorsList[r]);
				doc.setCharacterAttributes(cp - len, len,
								st, true);
				if (options.getProperty
						("speech").equals("true")) {
					r = rndgen.nextInt
						(congratulations.length); 
					voice.speak(congratulations[r] + "!");
				}
				// XXX: append a space into the text
				/*
				select(cp, cp);
				replaceSelection(" ");
				Style st2 = styles.addStyle(null,
						null);
				StyleConstants.setForeground(st2, Color.black);
				doc.setCharacterAttributes(cp, 2, st2, true);
				doc.setLogicalStyle(
					doc.getEndPosition().getOffset()-1,st2);
				*/
				// Select another word.
				selectRandomWord();
		}
	}


	// XXX: replace with a single function.
	private void initUserOptions() {
		findConfFile();
		loadOptions();
	}


	// Read options from the config file.
	private void loadOptions() {
		try {
			if (confFile.isFile() && confFile.canRead()) {
				FileInputStream fstrm = 
						new FileInputStream(confFile);
				options.load(fstrm);
				fstrm.close();
			} else {
				// Cannot get a config file; fall back
				// using the defaults.
				options = defaultProperties;
			}
		} catch (FileNotFoundException fnfe) {
			messagesDialog.showMessageDialog(getFrame(),
						"Cannot find options file.",
						"Warning",
						JOptionPane.WARNING_MESSAGE);
		} catch (IOException ioe) {
			messagesDialog.showMessageDialog(getFrame(),
						"Cannot read options file.",
						"Warning",
						JOptionPane.WARNING_MESSAGE);
		}
		// Time to set window's size.
		ww = (new Integer(
			options.getProperty("windowwidth", "640"))).intValue();
		wh = (new Integer(
			options.getProperty("windowheight", "480"))).intValue();
	}


	// XXX: a better algorithm is needed.
	private void findConfFile() {
		String fsep = System.getProperty("file.separator");
		File f;
		String userHome = System.getProperty("user.home");
		if (userHome != "") {
			// There is a user's home directory.
			File homeDir = new File(userHome + fsep + jrwDirName);
			if (!homeDir.isDirectory()) {
				// There isn't the .jrw dir in the home.
				homeDir.mkdirs();
				f = new File(userHome + fsep + jrwDirName +
						fsep + jrwConfFileName);
				saveOptions(defaultProperties, f);
				confFile = f;
				messagesDialog.showMessageDialog(getFrame(),
					"Configuration file created.",
					"Warning",
					JOptionPane.INFORMATION_MESSAGE);
				return;
			} else {
				// The directory ~/.jrw exists...
				f = new File(userHome + fsep + jrwDirName +
						fsep + jrwConfFileName);
				if (f.isFile() && f.canRead()) {
					// ...and even the configuration file!
					confFile = f;
					return;
				} else {
					// ...but not the configuration file!
					saveOptions(defaultProperties, f);
					confFile = f;
					messagesDialog.showMessageDialog
						(getFrame(),
						"Configuration file created.",
						"Warning",
						JOptionPane.INFORMATION_MESSAGE);
					return;
				}
			}
		}
		// Unable to find a user home directory.
		f = new File(jrwConfFileName);
		if (f.isFile() && f.canRead()) {
			// There is a configuration file in the
			// current directory.
			confFile = new File(jrwConfFileName);
			return;
		} else {
			// Try creating a configuration file in this dir.
			saveOptions(defaultProperties, f);
			confFile = f;
			messagesDialog.showMessageDialog(getFrame(),
					"Configuration file created.",
					"Warning",
					JOptionPane.INFORMATION_MESSAGE);
		}
		// Unable to get the configuration file.
		messagesDialog.showMessageDialog(getFrame(),
				"Unable to create a configuration file.",
				"Warning",
				JOptionPane.WARNING_MESSAGE);
	}


	// Save current set of options.
	private void saveOptions(Properties p, File f) {
		try {
			FileOutputStream fstrm = new FileOutputStream(f);
			p.store(fstrm, "Self generated file.  Do not edit!");
			fstrm.close();
		} catch (FileNotFoundException fnfe) {
			messagesDialog.showMessageDialog(getFrame(),
				"Unable to create the configuration file.",
				"Warning", JOptionPane.WARNING_MESSAGE);
		} catch (IOException ioe) {
			messagesDialog.showMessageDialog(getFrame(),
				"Unable to access the configuration file.",
				"Warning", JOptionPane.WARNING_MESSAGE);
		}
	}


	// Exclude unfair words.
	private boolean checkWord(String s) {
		if (s != null && !s.equals("") &&
				(s.indexOf(' ') == -1) &&
				(s.indexOf('\n') == -1) &&
				(s.indexOf('\t') == -1) &&
				(s.indexOf('\r') == -1) &&
				(s.indexOf('\f') == -1)) {
			return true;
		}
		return false;
	}


	// Initialize the words list.
	private void initWords() {
		readWordsList();
		updateConsideredWords();
		selectRandomWord();
	}


	// Read the file with the words list.
	private void readWordsList() {
		wordsList.clear();
		try {
			String txt;
			BufferedReader br = new BufferedReader(
					new FileReader(
					options.getProperty("wordsfile")));
			while (br.ready()) {
				txt = br.readLine().trim();
				if (checkWord(txt)) {
					wordsList.add(txt);
				}
			}
			br.close();
		} catch (FileNotFoundException fnfe) {
			messagesDialog.showMessageDialog(getFrame(),
				"Cannot find words list.",
				"Warning", JOptionPane.WARNING_MESSAGE);
		} catch (IOException ioe) {
			messagesDialog.showMessageDialog(getFrame(),
				"Unable to read words list.",
				"Warning", JOptionPane.WARNING_MESSAGE);
		}
	}


	//Update the considered set of words.
	private void updateConsideredWords() {
		consideredWords.clear();
		int i = 0;
		if (maxLength == -1) {
			consideredWords = (Vector) wordsList.clone();
		} else {
			while (i < wordsList.size()) {
				String elem = (String) wordsList.get(i);
				if (elem.length() <= maxLength) {
					consideredWords.add(elem);
				}
				i++;
			}
		}
	}
 
	// Pick a random words from the list.
	// XXX: be aware that other methods to select a word will be available.
	private void selectRandomWord() {
		if (consideredWords.size() > 0)  {
			int rn = rndgen.nextInt(consideredWords.size());
			currentWord = (String) consideredWords.get(rn);
		} else {
			currentWord = "";
			messagesDialog.showMessageDialog(getFrame(),
				"Cannot select any word.",
				"Warning", JOptionPane.WARNING_MESSAGE);

		}
		textlabel.setText(getResourceString("write-this") +
					currentWord.toUpperCase());
		if (consideredWords.size() > 0 &&
				options.getProperty("speech").equals("true")) {
			voice.speak(getResourceString("introduceword") +
					" \"" + currentWord + "\"");
		}
	}


	// Save the current set of words to the file.
	private void saveWordsToFile() {
		try {
			BufferedWriter bw = new BufferedWriter(
					new FileWriter(
					options.getProperty("wordsfile")));
			int i = 0;
			String txt = "";
			while (i < wordsList.size()) {
				txt = (String) wordsList.get(i);
				bw.write(txt, 0, txt.length());
				bw.newLine();
				i++;
			}
			bw.flush();
			bw.close();
		} catch (IOException ioe) {
			messagesDialog.showMessageDialog(getFrame(),
				"Unable to save words file.",
				"Warning", JOptionPane.WARNING_MESSAGE);
		}
	}


	// Show the options panel.
	private class ShowOptionsAction extends AbstractAction {

		ShowOptionsAction() {
			super("show-options");
		}
		ShowOptionsAction(String nm) {
			super(nm);
		}
		public void actionPerformed(ActionEvent e) {
			showOptionsDialog();
		}
	}


	// Exit via button or menu.
	private class ExitAction extends AbstractAction {

		ExitAction() {
			super("exit");
		}
		ExitAction(String nm) {
			super(nm);
		}
		public void actionPerformed(ActionEvent e) {
			int ret = askSaveDialog();
			switch (ret) {
				case 0:
					showSaveDialog();
					break;
				case 2:
					break;
				default:
					System.exit(0);
			}
		}
	}


	// A brand new board.
	private class NewAction extends AbstractAction {

		NewAction() {
			super("new");
		}
		NewAction(String nm) {
			super(nm);
		}
		public void actionPerformed(ActionEvent e) {
			int ret = askSaveDialog();
			switch (ret) {
				case 0:
					showSaveDialog();
					return;
				case 2:
					return;
			}
			StyleContext sc = new StyleContext();
			DefaultStyledDocument doc = new DefaultStyledDocument
									(sc);
			getEditor().initDocument(doc, sc);
			validate();
		}
	}


	// Change the current word.
	private class ChangeWordAction extends AbstractAction {

		ChangeWordAction() {
			super("change-word");
		}
		ChangeWordAction(String nm) {
			super(nm);
		}
		public void actionPerformed(ActionEvent e) {
			selectRandomWord();
		}
	}


	// Modify the words' list.
	private class ModifyAction extends AbstractAction {
		ModifyAction() {
			super("modify-words");
		}
		ModifyAction(String nm) {
			super(nm);
		}
		public void actionPerformed(ActionEvent e) {
			showModifyWords();
		}
	}


	// Open a saved document.
	private class OpenAction extends AbstractAction {
		OpenAction() {
			super("open");
		}
		OpenAction(String nm) {
			super(nm);
		}
		public void actionPerformed(ActionEvent e) {
			showOpenDialog();
		}
	}


	// Save the current document.
	private class SaveAction extends AbstractAction {
		SaveAction() {
			super("save");
		}		   
		SaveAction(String nm) {
			super(nm);
		}			   
		public void actionPerformed(ActionEvent e) {
			showSaveDialog();
		}		   
	}			   


	// Show the about dialog.
	private class AboutAction extends AbstractAction {
		AboutAction() {
			super("about");
		}
		AboutAction(String nm) {
			super(nm);
		}
		public void actionPerformed(ActionEvent e) {
			showAboutDialog();
		}
	}


	// Save this doc?
	// return: -1 if no modified, 0 if "ok", 1 if "no", 2 if "cancel".
	private int askSaveDialog() {
		int ret = -1;
		if (getEditor().wasModified()) {
			ret = messagesDialog.showConfirmDialog(getFrame(),
					"Unsaved board.\nSave it now?",
					"Save",
					JOptionPane.YES_NO_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE);
		}
		return ret;
	}


	// Modify the list of words.
	private void showModifyWords() {
		Frame frame = getFrame();
		if (wordsDialog == null) {
			JPanel p = new JPanel();
			wordsDialog = new JDialog(frame, "Words", true);
			Box wlbox = new Box(BoxLayout.Y_AXIS);
			JButton rmb = new JButton("Remove Word");
			final DefaultListModel dsm = new DefaultListModel();
			int i = 0;
			while (i < wordsList.size()) {
				dsm.addElement(wordsList.get(i));
				i++;
			}
			final JList jl = new JList(dsm);
			rmb.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					int ri = jl.getSelectedIndex();
					if (ri != -1) {
						dsm.remove(ri);
					}
				}
			});
			JButton rs = new JButton("Save");
			rs.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					int i = 0;
					wordsList.clear();
					while (i < dsm.getSize()) {
						wordsList.add(
							dsm.getElementAt(i));
						i++;
					}
					saveWordsToFile();
					updateConsideredWords();
					wordsDialog.setVisible(false);
					getEditor().requestFocus();
				}
			});
			JButton ok = new JButton("Ok");
			ok.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					int i = 0;
					wordsList.clear();
					while (i < dsm.getSize()) {
						wordsList.add(
							dsm.getElementAt(i));
						i++;
					}
					updateConsideredWords();
					wordsDialog.setVisible(false);
					getEditor().requestFocus();
				}
			});
			final JTextField tf = new JTextField(15);
			JPanel tfp = new JPanel();
			tfp.setLayout(new BorderLayout(1, 1));
			p.setLayout(new BorderLayout(10, 10));
			JButton addb = new JButton("Add this Word");
			addb.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					String wtxt = tf.getText().trim();
					if (checkWord(wtxt)) {
						dsm.addElement(wtxt);
						tf.setText("");
					} else {
						messagesDialog.showMessageDialog
						(getFrame(),
						"Bad character in the word.",
						"Warning",
						JOptionPane.WARNING_MESSAGE);
					}
				}
			});
			JScrollPane sp = new JScrollPane(jl);
			wlbox.add(sp);
			wlbox.add(rmb);
			tfp.add(tf, BorderLayout.WEST);
			tfp.add(addb, BorderLayout.EAST);
			Box buttonbox = new Box(BoxLayout.X_AXIS);
			buttonbox.add(ok);
			buttonbox.add(rs);
			wordsDialog.getContentPane().add(wlbox,
						BorderLayout.CENTER);
			wordsDialog.getContentPane().add(tfp,
						BorderLayout.NORTH);
			wordsDialog.getContentPane().add(buttonbox,
						BorderLayout.SOUTH);
			wordsDialog.pack();
		}
		wordsDialog.show();
	}


	// Just a moment to myself. ;-)
	private void showAboutDialog() {
		messagesDialog.showMessageDialog(getFrame(),
			JRW_NAME + " " + JRW_VERSION + "    \n" +
			"A small educational tool " +
			"under the " + JRW_LICENSE + " license.    \n\n" +
			"Copyright " + JRW_YEAR + " " + JRW_AUTHOR +
			"                      \n" + "<" + JRW_EMAIL + ">" +
			"   \n" + JRW_HOMEPAGE + "    ",
			"About " + JRW_NAME + "    ",
			JOptionPane.INFORMATION_MESSAGE);
	}


	// Options panel.
	private void showOptionsDialog() {
		Frame frame = getFrame();
		if (optionsDialog == null) {
			optionsDialog = new JDialog(frame, "Options", true);
			JPanel p = new JPanel();
			p.setLayout(new BorderLayout());
			JPanel centerp = new JPanel();
			centerp.setLayout(new BoxLayout(centerp,
						BoxLayout.X_AXIS));
			JPanel centerpr = new JPanel();
			centerpr.setLayout(new BoxLayout(centerpr,
						BoxLayout.Y_AXIS));
			JPanel centerpl = new JPanel();
			centerpl.setLayout(new BoxLayout(centerpl,
						BoxLayout.Y_AXIS));
			final JComboBox cp = new JComboBox();
			JLabel lab = new JLabel
				("Select maximum length of the words:");
			Box lenbox = new Box(BoxLayout.X_AXIS);
			cp.addItem("4");
			cp.addItem("5");
			cp.addItem("6");
			cp.addItem("7");
			cp.addItem("8");
			cp.addItem("unlimited");
			cp.setSelectedItem(options.getProperty("maxlength"));
			lenbox.add(lab);
			lenbox.add(cp);
			Box topbox = new Box(BoxLayout.Y_AXIS);
			topbox.add(lenbox);
			JLabel fsl = new JLabel("Font size:");
			final JComboBox fsv = new JComboBox();
			fsv.addItem("8");
			fsv.addItem("10");
			fsv.addItem("12");
			fsv.addItem("14");
			fsv.addItem("16");
			fsv.addItem("18");
			fsv.addItem("22");
			fsv.addItem("24");
			fsv.addItem("26");
			fsv.setSelectedItem(options.getProperty("fontsize"));
			final JCheckBox rw = new JCheckBox("Speech synthesis:");
			rw.setHorizontalTextPosition(JCheckBox.LEFT);
			rw.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (rw.isSelected()) {
						options.setProperty("speech",
							"true");
					} else {
						options.setProperty("speech",
							"false");
					}
				}
			});
			if (options.getProperty("speech").equals("true")) {
				rw.setSelected(true);
			} else {
				rw.setSelected(false);
			}
			Box fsbox = new Box(BoxLayout.X_AXIS);
			fsbox.add(fsl);
			fsbox.add(fsv);
			fsbox.add(rw);
			topbox.add(fsbox);
			p.add(topbox, BorderLayout.NORTH);
			JLabel whl = new JLabel("Window height:");
			final JTextField whv = new JTextField(4);
			whv.setText(options.getProperty("windowheight"));
			Box whbox = new Box(BoxLayout.X_AXIS);
			whbox.add(whl);
			whbox.add(whv);
			centerpl.add(whbox);
			JLabel wwl = new JLabel("Window width:");
			final JTextField wwv = new JTextField(4);
			wwv.setText(options.getProperty("windowwidth"));
			Box wwbox = new Box(BoxLayout.X_AXIS);
			wwbox.add(wwl);
			wwbox.add(wwv);
			centerpl.add(wwbox);
			JLabel ftl = new JLabel("Words file:");
			final JTextField ftv = new JTextField(20);
			ftv.setText(options.getProperty("wordsfile"));
			JButton ftb = new JButton("..");
			ftb.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					Frame frame = getFrame();
					if (fileChooser == null) {
						fileChooser=new JFileChooser();
					}	   
					fileChooser.showOpenDialog(frame);
					File f = fileChooser.getSelectedFile();
					if (f != null && f.exists() &&
								f.canRead()) {
						ftv.setText(
							f.getAbsolutePath());
					}
				}
			});
			Box ftbox = new Box(BoxLayout.X_AXIS);
			ftbox.add(ftl);
			ftbox.add(ftv);
			ftbox.add(ftb);
			centerpr.add(ftbox);
			centerp.add(centerpl);
			centerp.add(centerpr);
			p.add(centerp, BorderLayout.CENTER);
			JButton rs = new JButton("Save");
			rs.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					String ml = (String)
							cp.getSelectedItem();
					String whval = whv.getText();
					String wwval = wwv.getText();
					String ftval = ftv.getText();
					String fsval = (String)
							fsv.getSelectedItem();
					if (checkOptionsAndSave(ml, whval,
							wwval, ftval, fsval,
							rw.isSelected())) {
					}
					optionsDialog.setVisible(false);
					getEditor().requestFocus();
				}
			});
			JButton ok = new JButton("Ok");
			ok.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					String ml = (String)
						cp.getSelectedItem();
					String whval = whv.getText();
					String wwval = wwv.getText();
					String ftval = ftv.getText();
					String fsval = (String)
						fsv.getSelectedItem();
					if (checkOptions(ml, whval, wwval,
							ftval, fsval,
							rw.isSelected())) {
						optionsDialog.setVisible(false);
						getEditor().requestFocus();
					}
				}
			});
			Box buttonbox = new Box(BoxLayout.X_AXIS);
			buttonbox.add(ok);
			buttonbox.add(rs);
			p.add(buttonbox, BorderLayout.SOUTH);
			optionsDialog.getContentPane().add(p);
			optionsDialog.pack();
		}
		optionsDialog.show();
	}


	// Check options and - if all are valid - save.
	private boolean checkOptionsAndSave(String ml, String whval,
						String wwval, String ftval,
						String fsval, boolean sp) {
		boolean ret = checkOptions(ml, whval, wwval, ftval, fsval, sp);
		if (ret) {
			saveOptions(options, confFile);
		}
		return ret;
	}


	// Check the options.
	private boolean checkOptions(String ml, String whval, String wwval,
					String ftval, String fsval,
					boolean sp) {
		boolean ret = true;
		int max = -1;
		try {
			max = (new Integer(ml)).intValue();
		} catch (NumberFormatException nfe) {}
		maxLength = max;
		if (max == -1) {
			options.setProperty("maxlength", "unlimited");
		} else {
			options.setProperty("maxlength", ml);
		}
		updateConsideredWords();
		selectRandomWord();
		int wh = 0;
		int ww = 0;
		try {
			wh = (new Integer(whval)).intValue();
			ww = (new Integer(wwval)).intValue();
		} catch (NumberFormatException nfe) {
			messagesDialog.showMessageDialog(getFrame(),
				"Window's size must be numeric",
				"Warning", JOptionPane.WARNING_MESSAGE);
			ret = false;
		}
		if ((wh >= 100 && wh <= 1000) &&
			(ww >= 100 && ww <= 1000)) {
			options.setProperty("windowheight", whval);
			options.setProperty("windowwidth", wwval);
			getFrame().setSize(ww, wh);
		} else {
			messagesDialog.showMessageDialog(getFrame(),
				"Window's size out of limit (100~1000)",
				"Warning", JOptionPane.WARNING_MESSAGE);
			ret = false;
		}
		if (!ftval.equals(options.getProperty("wordsfile"))) {
			options.setProperty("wordsfile", ftval);
			initWords();
		}
		if (!fsval.equals(options.getProperty("fontsize"))) {
			options.setProperty("fontsize", fsval);
			getEditor().setFontSize();
		}
		if (sp) {
			options.setProperty("speech", "true");
		} else {
			options.setProperty("speech", "false");
		}
		return ret;
	}


	// Open a new doc.
	private void showOpenDialog() {
		int ret = askSaveDialog();
		switch (ret) {
			case 0:
				showSaveDialog();
				return;
			case 2:
				return;
		}
		Frame frame = getFrame();
		if (fileChooser == null) {
			fileChooser = new JFileChooser();
		}
		fileChooser.showOpenDialog(frame);
		File f = fileChooser.getSelectedFile();
		if (f != null && f.exists()) {
			try {
				FileInputStream fin = new FileInputStream(f);
				ObjectInputStream istrm =
						new ObjectInputStream(fin);
				DefaultStyledDocument doc =
						(DefaultStyledDocument)
						istrm.readObject();;
				StyleContext sc = new StyleContext();
				getEditor().initDocument(doc, sc);
				validate();
			} catch (IOException io) {
				messagesDialog.showMessageDialog(getFrame(),
					"Cannot read the file.",
					"Warning",
					JOptionPane.WARNING_MESSAGE);
			} catch (ClassNotFoundException cnf) {
				System.err.println("Class not found: " +
							cnf.getMessage());
			}
		} else {
			if (f != null) {
				messagesDialog.showMessageDialog(getFrame(),
						"Cannot read the file.",
						"Warning",
						JOptionPane.WARNING_MESSAGE);
			}
		}
	}


	// Saving the current doc.
	private void showSaveDialog() {
		Frame frame = getFrame();
		if (fileChooser == null) {
			fileChooser = new JFileChooser();
		}
		fileChooser.showSaveDialog(frame);
		File f = fileChooser.getSelectedFile();
		if (f == null || f.getName() == "") {
			return;
		}
		try {
			FileOutputStream fstrm = new FileOutputStream(f);
			ObjectOutput ostrm = new ObjectOutputStream(fstrm);
			ostrm.writeObject(getEditor().getDocument());
			ostrm.flush();
		} catch (IOException io) {
			messagesDialog.showMessageDialog(getFrame(),
				"Cannot write the file.",
				"Warning", JOptionPane.WARNING_MESSAGE);
			return;
		}
		getEditor().setModified(false);
	}


	// Return the current frame.
	private Frame getFrame() {
		for (Container p = getParent(); p != null; p = p.getParent()) {
			if (p instanceof Frame) {
				return (Frame) p;
			}
		}
		return null;
	}


	// This board.
	private static EditBoard getEditor() {
		return  editor;
	}


	// Fetch the command list for the editor.
	private Action[] getActions() {
		return TextAction.augmentList(editor.getActions(),
						defaultActions);
	}


	// Get the action for the given string.
	private Action getAction(String cmd) {
		return (Action) commands.get(cmd);
	}
	

	// Buid the menus.
	private void createMenubar() {
		menubar = new JMenuBar();
		String[] menuKeys = tokenize(getResourceString("menubar"));
		for (int i = 0; i < menuKeys.length; i++) {
			JMenu m = createMenu(menuKeys[i]);
			menubar.add(m);
		}
	}


	// Create a menu.
	private JMenu createMenu(String key) {
		String[] itemKeys = tokenize(getResourceString(key));
		JMenu menu = new JMenu(getResourceString(key + "Label"));
		for (int i = 0; i < itemKeys.length; i++) {
			if (itemKeys[i].equals("-")) {
				menu.addSeparator();
			} else {
				JMenuItem mi = createMenuItem(itemKeys[i]);
				menu.add(mi);
			}
		}
		return menu;
	}


	// Create an entry for a menu.
	private JMenuItem createMenuItem(String cmd) {
		JMenuItem mi = new JMenuItem(getResourceString(cmd + "Label"));
		String astr = getResourceString(cmd + "Action");
		if (astr == null) {
			astr = cmd;
		}
		mi.setActionCommand(astr);
		Action a = getAction(astr);
		if (a != null) {
			mi.addActionListener(a);
		} else {
			mi.setEnabled(false);
		}
		return mi;
	}


	// Create the buttons bar.
	private Component createToolbar() {
		toolbar = new JToolBar();
		String[] toolKeys = tokenize(getResourceString("toolbar"));
		for (int i = 0; i < toolKeys.length; i++) {
			if (toolKeys[i].equals("-")) {
				toolbar.add(Box.createHorizontalStrut(5));
			} else {
				toolbar.add(createToolbarButton(toolKeys[i]));
			}
		}
		toolbar.add(Box.createHorizontalGlue());
		return toolbar;
	}


	// Create a single button.
	private JButton createToolbarButton(String key) {
		URL url = getResource(key + "Image");
		JButton b = new JButton(new ImageIcon(url)) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		b.setRequestFocusEnabled(false);
		b.setMargin(new Insets(1, 1, 1, 1));

		String astr = getResourceString(key + "Action");
		if (astr == null) {
			astr = key;
		}
		Action a = getAction(astr);
		if (a != null) {
			b.setActionCommand(astr);
			b.addActionListener(a);
		} else {
			b.setEnabled(false);
		}
		String tip = getResourceString(key + "Tooltip");
		if (tip != null) {
			b.setToolTipText(tip);
		}
		return b;
	}


	// Split a string into tokens.
	private String[] tokenize(String input) {
		StringTokenizer t = new StringTokenizer(input);
		String[] ret = new String[t.countTokens()];
		int i = 0;
		while (t.hasMoreTokens()) {
			ret[i] = t.nextToken();
			i++;
		}
		return ret;
	}

	// Populate the greetings list of strings.
	private void getCongratulations() {
		StringTokenizer t = new StringTokenizer(
				getResourceString("congratulations"), ";");
		congratulations = new String[t.countTokens()];
		int i = 0;
		while (t.hasMoreTokens()) {
			congratulations[i] = t.nextToken();
			i++;
		}
	}

	// Read a resource.
	private String getResourceString(String nm) {
		String str;
		try {
			str = resources.getString(nm);
		} catch (MissingResourceException mre) {
			str = null;
		}
		return str;
	}


	// Return a path.
	private URL getResource(String key) {
		String name = getResourceString(key);
		if (name != null) {
			URL url = this.getClass().getResource(name);
			return url;
		}
		return null;
	}

}


